library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package record_types_pkg is

    subtype ip_address_type is std_logic_vector(31 downto 0);
    subtype mac_address_type is std_logic_vector(47 downto 0);

    -- A record type to represent an entry in the ARP (address resolution protocol) table
    type arp_entry_type is record
        ip: ip_address_type;
        mac: mac_address_type;
        is_valid: boolean;
    end record;

    subtype coordinate_type is natural range 0 to 2047;

    -- A record type to represent a point on the screen
    type point_type is record
        x: coordinate_type;
        y: coordinate_type;
    end record;

    -- A record type to represent a rectangle by two of its corners
    type rectangle is record
        top_left: point_type;
        bottom_right: point_type;
    end record;

    -- An unconstrained record type. Objects of this type can have different
    -- numbers of bits, and their sizes must be set at the object declaration.
    type color_type is record
        red: unsigned;
        green: unsigned;
        blue: unsigned;
    end record;

    -- A constrained subtype of a record type. Objects of this type must have
    -- the number of bits given by the subtype constraints.
    subtype color_24bit_type is color_type(
        red(7 downto 0), green(7 downto 0), blue(7 downto 0)
    );

    -- An instance of a constrained record subtype. It is not necessary to
    -- specify the numbers of bits because they were set at the subtype declaration.
    constant BLACK_24_BIT: color_24bit_type := (others => (others => '0'));

    -- An instance of an unconstrained record type. We nees to specify the
    -- numbers of bits because they were not set at the type declaration.
    constant BLACK_16_BIT: color_type(red(4 downto 0), green(5 downto 0), blue(4 downto 0)) :=
        (others => (others => '0'));

end;