-- Types and operations for working with boolean matrices. In a boolean matrix,
-- elements are of type bit, and operations are defined over the finite field
-- GF(2).
package bit_matrix_pkg is
    type matrix_type is array (natural range <>, natural range <>) of bit;

    function matrix_column(m: matrix_type; c: integer) return bit_vector;
    function matrix_row(m: matrix_type; r: integer) return bit_vector;
    function height(matrix: matrix_type) return integer;
    function width(matrix: matrix_type) return integer;
end;

package body bit_matrix_pkg is

    function height(matrix: matrix_type) return integer is
    begin
        return matrix'length(1);
    end;

    function width(matrix: matrix_type) return integer is
    begin
        return matrix'length(2);
    end;

    -- Extract a matrix column and return it as a vector
    function matrix_column(m: matrix_type; c: integer) return bit_vector is
        variable retVect: bit_vector(m'range);
    begin
        for i in retVect'range loop
            retVect(i) := m(i, c);
        end loop;
        return retVect;
    end;

    -- Extract a matrix row and return it as a vector
    function matrix_row(m: matrix_type; r: integer) return bit_vector is
        variable retVect: bit_vector(m'range(2));
    begin
        for j in retVect'range loop
            retVect(j) := m(r, j);
        end loop;
        return retVect;
    end;

    function "*"(A, B: matrix_type) return matrix_type is
        variable C: matrix_type(1 to width(A), 1 to height(B));
        variable A_row: bit_vector(A'range(2));
        variable B_col: bit_vector(B'range(1));
    begin
        -- Product A x B is defined only if the number of
        -- columns in A equals the number of rows in B
        assert width(A) = height(B)
            report "Invalid matrix dimensions in '*'"
            severity failure;

        for i in C'range(1) loop
            A_row := matrix_row(A, i);
            for j in C'range(2) loop
                B_col := matrix_column(B, j);
                C(i, j) := xor (A_row and B_col);
            end loop;
        end loop;

        return C;
    end;

end;
