library ieee;
use ieee.std_logic_1164.all;
use work.assert_pkg.all;
use work.print_pkg.all;

entity vending_machine_fsm_tb is
end;

architecture testbench of vending_machine_fsm_tb is
    signal clock: std_ulogic := '0';
    signal reset: std_ulogic;
    signal nickel_in: std_ulogic;
    signal dime_in: std_ulogic;
    signal dispense: std_ulogic;
    signal amount: std_ulogic_vector(3 downto 0);

    constant CLOCK_PERIOD: time := 20 ns;
    constant COMBINATIONAL_DELAY: time := 6 ns;
begin
    duv: entity work.vending_machine_fsm port map (
        clock => clock, reset => reset, nickel_in => nickel_in,
        dime_in => dime_in, dispense => dispense, amount => amount);

    clock <= not clock after CLOCK_PERIOD / 2;

    test_sequencer: process
        type test_data_type is record
            nickel: std_logic;
            dime: std_logic;
            dispense: std_logic;
            amount: std_logic_vector(3 downto 0);
        end record;

        type test_data_array_type is array (natural range <>) of test_data_type;

        procedure wait_clock_edge is begin
            wait until rising_edge(clock);
            wait for COMBINATIONAL_DELAY;
        end;

        procedure reset_with_default_inputs is begin
            nickel_in <= '0';
            dime_in <= '0';
            reset <= '0';
            wait_clock_edge;
        end;

        procedure test_reset is begin
            print("test_reset");
            reset_with_default_inputs;
            assert_eq(dispense, '0', "dispense is 0 following reset");
            assert_eq(amount, "0000", "amount is 0000 following reset");
        end;

        procedure test_input_sequence(name: string; seq: test_data_array_type) is
        begin
            print("test_input_sequence: " & name);
            reset_with_default_inputs;

            for i in seq'range loop
                nickel_in <= seq(i).nickel;
                dime_in <= seq(i).dime;
                wait_clock_edge;
                assert_eq(
                    dispense, seq(i).dispense,
                    "dispense is " & to_string(seq(i).dispense));
                assert_eq(
                    amount, seq(i).amount,
                    "amount is " & to_string(seq(i).amount));
            end loop;
        end;

    begin
        test_reset;

        test_input_sequence( name => "nickel, nickel, nickel", seq => (
            (nickel => '1', dime => '0', dispense => '0', amount => 4d"05"),
            (nickel => '1', dime => '0', dispense => '0', amount => 4d"10"),
            (nickel => '1', dime => '0', dispense => '1', amount => 4d"15"),
            (nickel => '0', dime => '0', dispense => '0', amount => 4d"00")
        ) );

        test_input_sequence( name => "nickel, dime", seq => (
            (nickel => '1', dime => '0', dispense => '0', amount => 4d"05"),
            (nickel => '0', dime => '1', dispense => '1', amount => 4d"15"),
            (nickel => '0', dime => '0', dispense => '0', amount => 4d"00")
        ) );

        print("End of testbench. All tests completed successfully.");
        std.env.finish;
    end process;
end;