library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.assert_pkg.all;
use work.print_pkg.all;

entity vending_machine_fsm_tb is
end;

architecture testbench of vending_machine_fsm_tb is
    signal clock: std_ulogic := '0';
    signal reset: std_ulogic := '0';
    signal nickel_in: std_ulogic := '0';
    signal dime_in: std_ulogic := '0';
    signal dispense: std_ulogic;
    signal amount: std_ulogic_vector(3 downto 0);

    constant CLOCK_PERIOD: time := 20 ns;
    constant COMBINATIONAL_DELAY: time := 6 ns;

    type coins_transaction_type is record
        nickels: std_logic_vector(1 to 5);
        dimes: std_logic_vector(1 to 5);
    end record;

    signal transaction_data: coins_transaction_type;
    signal transaction_done: boolean;

    procedure wait_clock_edge is begin
        wait until rising_edge(clock);
        wait for COMBINATIONAL_DELAY;
    end;
begin
    duv: entity work.vending_machine_fsm port map (
        clock => clock, reset => reset, nickel_in => nickel_in,
        dime_in => dime_in, dispense => dispense, amount => amount);

    clock <= not clock after CLOCK_PERIOD / 2;
    reset <= '1', '0' after 50 ns;

    stimuli_generator: process
        variable seq: std_logic_vector(1 to 10);
    begin
        wait until not reset;

        -- Generate all possible 10-bit sequences
        for i in 0 to 2**10-1 loop
            seq := std_logic_vector(to_unsigned(i, 10));
            -- Break down each 10-bit sequence into sequences of nickels and dimes
            transaction_data <= (nickels => seq(1 to 5), dimes => seq(6 to 10));
            wait on transaction_done'transaction;
        end loop;

        print("End of testbench. All tests completed successfully.");
        std.env.finish;
    end process;

    driver: process begin
        wait on transaction_data'transaction;

        for i in 1 to 5 loop
            nickel_in <= transaction_data.nickels(i);
            dime_in <= transaction_data.dimes(i);
            wait_clock_edge;
        end loop;

        transaction_done <= true;
    end process;

    response_checker: process
        variable amount_predicted, amount_duv: natural range 0 to 15;
        variable dispense_predicted: std_logic;
    begin
        wait_clock_edge;

        -- Predictor for the 'amount' output
        if reset = '1' or amount_predicted = 15 then
            amount_predicted := 0;
        elsif dime_in then
            amount_predicted := minimum(amount_predicted + 10, 15);
        elsif nickel_in then
            amount_predicted := amount_predicted + 5;
        end if;

        -- Predictor for the 'dispense' output
        amount_duv := to_integer(unsigned(testbench.amount));
        dispense_predicted := '1' when amount_duv = 15 else '0';

        -- Comparator - assert that DUV outputs and predictor outputs match
        assert_eq(amount_predicted, amount_duv,
            "amount is as expected: " & to_string(amount_duv));
        assert_eq(dispense_predicted, dispense,
            "dispense is as expected: " & to_string(dispense));
    end process;
end;